@@ -2,9 +2,11 @@ import functools
 
 import numpy as np
 from numpy import ma as ma
+from pycce.utilities import gen_state_list
+from numba.typed import List
 
 
-def generate_bath_state(bath, nbstates, seed=None, fixstates=None, parallel=False):
+def generate_bath_state(bath, nbstates, seed=None, parallel=False):
     r"""
     Generator of the random *pure* :math:`\hat{I}_z` bath eigenstates.
 
@@ -12,14 +14,15 @@ def generate_bath_state(bath, nbstates, seed=None, fixstates=None, parallel=Fals
         bath (BathArray): Array of bath spins.
         nbstates (int): Number of random bath states to generate.
         seed (int): Optional. Seed for RNG.
-        fixstates (dict): Optional. dict of which bath states to fix. Each key is the index of bath spin,
-            value - fixed :math:`\hat{I}_z` projection of the mixed state of nuclear spin.
+        parallel (bool): True if run in parallel mode. Default False.
 
     Yields:
-        ndarray: Array of ``shape = len(bath)`` containing z-projections of the bath spins states.
+        List: list of the pure bath spin state vectors.
     """
     rgen = np.random.default_rng(seed)
     rank = 0
+    comm = None
+
     if parallel:
         try:
             import mpi4py
@@ -30,21 +33,26 @@ def generate_bath_state(bath, nbstates, seed=None, fixstates=None, parallel=Fals
             print('Parallel failed: mpi4py is not found. Running serial')
             parallel = False
 
+    dimensions = np.empty(bath.shape, dtype=np.int32)
+
     for _ in range(nbstates):
         bath_state = np.empty(bath.shape, dtype=np.float64)
+
         if rank == 0:
-            for n in bath.types:
+
+            for n in np.unique(bath.N):
                 s = bath.types[n].s
-                snumber = int(round(2 * s + 1))
-                mask = bath['N'] == n
+                snumber = np.int32(round(2 * s + 1))
+                mask = bath.N == n
                 bath_state[mask] = rgen.integers(snumber, size=np.count_nonzero(mask)) - s
 
-            if fixstates is not None:
-                for fs in fixstates:
-                    bath_state[fs] = fixstates[fs]
+                dimensions[mask] = snumber
 
         if parallel:
             comm.Bcast(bath_state, root=0)
+            comm.Bcast(dimensions, root=0)
+
+        bath_state = gen_state_list(bath_state, dimensions)
 
         yield bath_state
 
@@ -65,7 +73,10 @@ def monte_carlo_method_decorator(func):
                 print('Parallel failed: mpi4py is not found. Running serial')
                 self.parallel_states = False
 
-        if self.masked:
+        if isinstance(self.masked, bool) and not self.masked:
+            self.masked = None
+
+        if self.masked is not None:
             divider = 0
         else:
             root_divider = self.nbstates
@@ -88,20 +99,24 @@ def monte_carlo_method_decorator(func):
 
         total = 0j
 
-        for bath_state in generate_bath_state(self.bath, nbstates, seed=seed,
-                                              fixstates=self.fixstates, parallel=self.parallel):
-            self.bath_state = bath_state
-            self.projected_bath_state = bath_state
+        his = self.bath.state.has_state.copy()  # have_initial_state
+        for bath_state in generate_bath_state(self.bath[~his], nbstates, seed=seed, parallel=self.parallel):
+
+            self.bath.state[~his] = bath_state
             result = func(self, *args, **kwargs)
 
-            if self.masked:
-                reshaped_result = np.abs(result).reshape(result.shape[0], -1)
-                proper = np.all(reshaped_result <= reshaped_result[0] * 1.01, axis=(-1))
+            if self.masked is not None:
+                if isinstance(self.masked, bool):
+                    self.masked = np.abs(result[0] * 1.01)
+
+                proper = np.abs(result) <= self.masked
                 divider += proper.astype(np.int32)
                 result[~proper] = 0.
 
             total += result
 
+        self.bath.state[~his] = None
+
         if self.parallel_states:
             if rank == 0:
                 result_shape = total.shape
@@ -117,7 +132,7 @@ def monte_carlo_method_decorator(func):
             root_result = ma.array(np.zeros(result_shape), dtype=np.complex128)
             comm.Allreduce(total, root_result, MPI.SUM)
 
-            if self.masked:
+            if self.masked is not None:
                 if rank == 0:
                     divider_shape = divider.shape
                 else:
@@ -130,14 +145,14 @@ def monte_carlo_method_decorator(func):
 
         else:
             root_result = total
-            if self.masked:
+            if self.masked is not None:
                 root_divider = divider
 
         root_result = ma.array(root_result, fill_value=0, dtype=np.complex128)
 
-        if self.masked:
+        if self.masked is not None:
             root_result[root_divider == 0] = ma.masked
-            root_divider = root_divider.reshape(root_result.shape[0], *[1] * len(root_result.shape[1:]))
+            # root_divider = root_divider.reshape(root_result.shape[0], *[1] * len(root_result.shape[1:]))
 
         root_result /= root_divider
 
